# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


INDENTED_SCRIPT = '''
    echo '
    a
    '
'''


SCRIPT_WITH_HEREDOCS = r'''
 b=X
 cat <<EOF
  'a'
  $b
 EOF
 cat <<\EOF
  'a'
  $b
 EOF
'''


@fingertip.transient
def main(m):
    with m:
        assert m.exec('true')
        assert m('true')
        assert m('sh -c true')
        assert m('sh', '-c', 'true', shell=False)

        assert m(INDENTED_SCRIPT).out == '\na\n\n'
        assert m(INDENTED_SCRIPT, dedent=False).out == '\n    a\n    \n'

        assert m(SCRIPT_WITH_HEREDOCS).out == " 'a'\n X\n 'a'\n $b\n"

        assert m.exec('echo', 'x', 'y', 'z').out == b'x y z\n'
        assert m('echo x y z').out == 'x y z\n'
        assert m('echo x y z', decode=False).out == b'x y z\n'

        assert m.exec('echo', '""').out == b'""\n'
        assert m('echo', '""', shell=False).out == '""\n'
        assert m('echo ""').out == '\n'

        # TODO: these won't work for SSH and ash
        # assert m.exec('echo', "''").out == b"''\n"
        # assert m('echo', "''", shell=False).out == "''\n"
        # assert m("echo ''").out == '\n'

        t = m('true')
        assert t

        f = m('false', check=False)
        assert not f
        ret, out, err = f
        assert (ret, out, err) == (1, '', '')

        e = m('echo a; echo e >/dev/stderr; echo b; exit 3', check=False)
        assert e.retcode == 3
        assert e.out == 'a\nb\n'
        assert e.err == 'e\n'

        e = m('''
          yes out1 | head -n 5000
          yes err1 | head -n 5000 > /dev/stderr
          yes out2 | head -n 5000
          yes err2 | head -n 5000 > /dev/stderr
        ''')
        assert e.out == 'out1\n' * 5000 + 'out2\n' * 5000
        assert e.err == 'err1\n' * 5000 + 'err2\n' * 5000



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

from fingertip.plugins.script._run_and_test import run, test

# debug is in fingertip.plugins.script.debug,
# but I'm not importing to minimize dependencies / rebuilds



"""
Helper functions for fingertep: inter-process locking.
"""

import collections
import threading

import fasteners


threadlocks = collections.defaultdict(threading.Lock)
threadlocks_lock = threading.Lock()


class LockTimeout(RuntimeError):
    pass


class Lock:
    def __init__(self, path, timeout=None):
        self.timeout = timeout
        # InterProcessLock does not work across threads
        self._plock = fasteners.process_lock.InterProcessLock(path)
        with threadlocks_lock:
            self._tlock = threadlocks[path]
        self._plock_acquired, self._tlock_acquired = False, False

    def __enter__(self):
        if not self._plock.acquire(timeout=self.timeout):
            raise LockTimeout('timeout acquiring interprocess lock')
        self._plock_acquired = True
        if not self._tlock.acquire(timeout=(self.timeout or -1)):
            raise LockTimeout('timeout acquiring interthread lock')
        self._tlock_acquired = True
        return self

    def __exit__(self, *_):
        if self._tlock_acquired:
            self._tlock.release()
            self._tlock_acquired = False
        if self._plock_acquired:
            self._plock.release()
            self._plock_acquired = False


# Python 3.7 has `contextlib.nullcontext`
class NoLock:
    def __enter__(self):
        pass

    def __exit__(self, *_):
        pass



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions for fingertip: working with self-destructing tempfiles.
"""

# TODO: replace with python3-tempdir

import atexit
import os
import random
import shutil
import signal
import string
import tempfile

from fingertip.util import log, units


AUTOREMOVE_PREFIX = 'tmp.fingertip'


# TODO: hacky and unclean
def terminate_child(num, frame):
    print('caught SIGTERM, cleaning up...')
    raise SystemExit('SIGTERM')  # fire atexit hooks


assert signal.signal(signal.SIGTERM, terminate_child) == signal.SIG_DFL


def random_chars(k=6):
    return ''.join(random.choices(string.ascii_letters, k=k))


def suffix_non_existing(base, k=6):  # racy
    while True:
        suffix = random_chars(k)
        if not os.path.exists(base + suffix):
            return base + suffix


def unique_filename(dstdir=None, hint=''):
    dstdir = dstdir or tempfile.gettempdir()
    hint = hint if len(hint) <= 20 else hint[:20-2] + '..'
    return suffix_non_existing(os.path.join(dstdir, f'.{hint}.'))


def unique_file(dstdir=None, hint='', create=False):
    tgt = unique_filename(dstdir=dstdir, hint=hint)
    if create:
        open(tgt, 'w').close()
    return tgt


def unique_dir(dstdir=None, hint=''):
    tgt = unique_filename(dstdir=dstdir, hint=hint)
    os.mkdir(tgt)
    return tgt


def remove(*paths):
    for path in paths:
        assert os.path.basename(path).startswith('.' + AUTOREMOVE_PREFIX)
        try:
            if not os.path.isdir(path) or os.path.islink(path):
                try:
                    os.unlink(path)
                except FileNotFoundError:
                    pass
            else:
                shutil.rmtree(path, ignore_errors=True)
        except Exception as e:
            log.warning(f'cleanup error for {path}: {e}')


def disappearing(path):
    assert os.path.basename(path).startswith('.' + AUTOREMOVE_PREFIX)
    atexit.register(remove, path)
    return path


def disappearing_file(dstdir=None, hint='', create=False):
    hint = f'{AUTOREMOVE_PREFIX}.{hint}' if hint else AUTOREMOVE_PREFIX
    return disappearing(unique_file(dstdir=dstdir, hint=hint, create=create))


def disappearing_dir(dstdir=None, hint=''):
    hint = f'{AUTOREMOVE_PREFIX}.{hint}' if hint else AUTOREMOVE_PREFIX
    return disappearing(unique_dir(dstdir=dstdir, hint=hint))


def has_space(how_much='2G', reserve_fraction=.5, where=None):
    where = where or tempfile.gettempdir()
    how_much = units.parse_binary(how_much)
    total, _, free = shutil.disk_usage(where)
    if not free >= how_much:
        log.warning(f'{where} does not have {how_much} of free space')
    if not free >= total * reserve_fraction:
        log.warning(f'{where} is {int((1 - reserve_fraction) * 100)}% full')
    return free >= how_much and free >= total * reserve_fraction



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

TIME = {'s': 1, 'm': 60, 'h': 3600, 'd': 24 * 3600, 'w': 24 * 3600 * 7}
BINARY = {'K': 2**10, 'M': 2**20, 'G': 2**30, 'T': 2**40, 'P': 2**50}


def parse_time_interval(interval):
    if isinstance(interval, str) and interval[-1] in TIME:
        return float(interval[:-1]) * TIME[interval[-1]]
    try:
        return float(interval)
    except ValueError:
        raise ValueError(f'Cannot parse time interval {interval}')


def parse_binary(value):
    if isinstance(value, str) and value[-1] in BINARY:
        return int(value[:-1]) * BINARY[value[-1]]
    try:
        return int(value)
    except ValueError:
        raise ValueError(f'Cannot parse binary-suffixed value {value}')


def binary(value):
    value = int(value)
    if not value:
        return '0'
    r = str(value)
    for suffix, suffix_value in BINARY.items():
        if value // suffix_value * suffix_value == value:
            r = f'{int(value / suffix_value)}{suffix}'
    return r



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

"""
`fingertip saviour <saviour-config.yml>` takes a config file and
locally mirrors whatever is specified there
(usually package repositories and GIT repositories)
in a way that's usable with local and non-local fingertip instances.
Example config is available as saviour.example.yml

The result is going to be available at
~/.cache/fingertip/saviour/http://whatever/you/have/mirrored.
Exporting it with an HTTP server as, e.g., http://cachy/saviour
(i.e., as http://cachy/saviour/http://whatever/you/have/mirrored)
will make it possible to make some other fingertip use this cache
with http_cache and git_cache
by setting FINGERTIP_SAVIOUR='cachy/saviour'.

For real applications you probably want
'local,cachy/saviour,direct' or 'local,cachy/saviour,cached+direct'.

Requires reflinking support (btrfs or XFS CoW).
Requires git and rsync, also dnf for dnf method.
"""

import collections
import fnmatch
import logging
import os
import shutil
import subprocess
import textwrap
import traceback

import git
import ruamel.yaml

import fingertip.expiration
import fingertip.machine
from fingertip.util import lock, log, path, reflink, temp


def _remove(tgt):
    assert tgt.startswith(path.SAVIOUR)
    if not os.path.exists(tgt):
        pass
    elif os.path.isdir(tgt):
        shutil.rmtree(tgt)
    else:
        os.unlink(tgt)


# base is a previous consistent snapshot, dst is a WIP directory
def method_rsync(log, src, base, dst, options=[], excludes=[]):
    run = log.pipe_powered(subprocess.run,
                           stdout=logging.INFO, stderr=logging.WARNING)
    run(['rsync', '-rvt', '--partial', '--del', '--delete-excluded'] +
        (['--copy-dest', base] if os.path.isdir(base) else []) +
        sum([['--exclude', e] for e in excludes], []) + options +
        [src, dst], check=True)


def method_git(log, src, base, dst):
    fingertip.util.log.info(f'removing {dst}...')
    _remove(dst)
    fingertip.util.log.info(f'cloning {src}...')
    r = git.Repo.clone_from(src, dst, mirror=True,
                            dissociate=True, reference_if_able=base)
    r.git.update_server_info()
    with open(os.path.join(dst, 'hooks/post-update.sample'), 'w') as f:
        f.write('#!/usr/bin/sh\nexec git update-server-info')


def method_reposync(log, src, base, dst,
                    arches=['noarch', 'x86_64'], source='auto',
                    metadata='download', options=[]):
    if source == 'auto':
        source = '/source' in src or '/SRPM' in src
    repo_desc_for_mirroring = textwrap.dedent(f'''
        [repo]
        baseurl = {src}
        name = repo
        enabled = 1
        gpgcheck = 0
    ''')
    repodir = temp.disappearing_dir()
    with open(os.path.join(repodir, f'whatever.repo'), 'w') as f:
        f.write(repo_desc_for_mirroring)
    run = log.pipe_powered(subprocess.run,
                           stdout=logging.INFO, stderr=logging.WARNING)
    run(['dnf', f'--setopt=reposdir={repodir}', 'reposync', '--norepopath',
         f'--download-path={dst}', '--repoid=repo',
         '--delete', '--remote-time'] +
        [f'--arch={arch}' for arch in arches] +
        (['--download-metadata'] if not metadata != 'generate' else []) +
        (['--source'] if source else []) +
        options,
        check=True)
    run = log.pipe_powered(subprocess.run,  # either too silent or too noisy =/
                           stdout=logging.INFO, stderr=logging.INFO)
    createrepo_c_options = ['-v', '--error-exit-val', '--ignore-lock']
    if metadata == 'regenerate':
        log.info('regenerating metadata...')
        run(['createrepo_c'] + createrepo_c_options + ['--update', dst],
            check=True)
    elif metadata == 'generate':
        log.info('generating metadata from scratch...')
        run(['createrepo_c'] + createrepo_c_options + [dst], check=True)


def method_command(log, src, base, dst, command='false', reuse=True):
    if not reuse:
        fingertip.util.log.info(f'removing {dst}...')
        _remove(dst)
    env = os.environ.copy()
    env['SRC'], env['BASE'], env['DST'] = src, base, dst
    run = log.pipe_powered(subprocess.run,
                           stdout=logging.INFO, stderr=logging.WARNING)
    log.info(command.replace('$SRC', src).replace('$BASE', base)
                    .replace('$DST', dst))
    run(command, shell=True, cwd=os.path.dirname(dst), env=env,
        check=True)


@fingertip.transient
def main(*args):
    if len(args) >= 1:
        subcmd, *args = args
        if subcmd == 'mirror':
            return mirror(*args)
        if subcmd == 'deduplicate' and not args:
            return deduplicate(log.Sublogger('deduplicate'))
    log.error('usage: ')
    log.error('    fingertip saviour mirror <config-file> [<what-to-mirror>]')
    log.error('    fingertip saviour deduplicate')
    raise SystemExit()


def _symlink(src, dst):
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    tmp = dst + '-tmp'
    if not os.path.lexists(tmp):
        os.symlink(src, tmp)
    os.replace(tmp, dst)


def mirror(config, *what_to_mirror, deduplicate=True):
    total_failures = []
    failures = collections.defaultdict(list)

    with open(config) as f:
        config = ruamel.yaml.YAML(typ='safe').load(f)
    hows, whats = config['how'], config['what']
    if not what_to_mirror:
        what_to_mirror = whats.keys()
    else:
        what_to_mirror = [k for k in whats.keys()
                          if any((fnmatch.fnmatch(k, req)
                                  for req in what_to_mirror))]

    for resource_name in what_to_mirror or whats.keys():
        s = whats[resource_name]
        log.debug(f'processing {resource_name}...')

        if s is None:
            how, suffix = resource_name, ''
        elif '/' in s:
            how, suffix = s.split('/', 1)
            suffix = '/' + suffix
        else:
            how, suffix = s, ''

        try:
            how = hows[how]
        except KeyError:
            log.error(f'missing how section on {how}')
            raise SystemExit()

        url = how['url'] + suffix
        method = how['method']
        sources = (how['sources'] if 'sources' in how else [how['url']])
        sources = [s + suffix for s in sources]
        extra_args = {k: v for k, v in how.items()
                      if k not in ('url', 'sources', 'method')}

        if f'method_{method}' not in globals():
            log.error(f'unsupported method {method}')
            raise SystemExit()

        meth = globals()[f'method_{method}']
        symlink = path.saviour(url.rstrip('/'))
        # usually symlink points to data, but while we're working on it,
        # it temporarily points to a consistent snapshot of it named `snap`
        data = path.saviour('_', resource_name, 'data')
        snap = path.saviour('_', resource_name, 'snap')
        temp = path.saviour('_', resource_name, 'temp')
        lockfile = path.saviour('_', resource_name) + '-lock'
        assert data.startswith(path.SAVIOUR)
        assert snap.startswith(path.SAVIOUR)
        assert temp.startswith(path.SAVIOUR)

        sublog = log.Sublogger(f'{method} {resource_name}')
        sublog.info('locking...')
        with lock.Lock(lockfile):
            os.makedirs(os.path.dirname(snap), exist_ok=True)

            if os.path.exists(temp):
                sublog.info('removing stale temp...')
                _remove(temp)
            if os.path.exists(symlink):  # it's already published
                if os.path.exists(data) and not os.path.exists(snap):
                    # `data` is present and is the best we have to publish
                    sublog.info('snapshotting...')
                    reflink.always(data, temp, preserve=True)
                    os.rename(temp, snap)
                if os.path.exists(snap):
                    # link to a consistent snapshot while we work on `data`
                    _symlink(snap, symlink)

            for source in sources:
                sublog.info(f'trying {source}...')
                try:
                    meth(sublog, source, snap, data, **extra_args)
                    assert os.path.exists(data)
                    break
                except Exception as _:
                    traceback.print_exc()
                    failures[resource_name].append(source)
                    fingertip.util.log.warning(f'failed to mirror {source}')

            if len(failures[resource_name]) == len(sources):
                sublog.error(f'failed to mirror '
                             f'from all {len(sources)} sources')
                total_failures.append(resource_name)
                continue

            _symlink(data, symlink)
            if os.path.exists(snap):
                os.rename(snap, temp)  # move it out the way asap
                sublog.info('removing now obsolete snapshot...')
                _remove(temp)

            if deduplicate:
                try:
                    _deduplicate(sublog, resource_name, timeout=1)
                except lock.LockTimeout:
                    log.warning('skipped deduplication, db was locked')
    if total_failures:
        fingertip.util.log.error(f'failed: {", ".join(total_failures)}')
        raise SystemExit()
    log.info('saviour has completed mirroring')


def deduplicate(log, *subpath, timeout=None):
    log.info('locking the deduplication db...')
    with lock.Lock(path.saviour('.duperemove.hashfile-lock'), timeout=timeout):
        log.info('deduplicating...')
        run = log.pipe_powered(subprocess.run,
                               stdout=logging.INFO, stderr=logging.WARNING)
        r = run(['duperemove',
                 '--hashfile', path.saviour('.duperemove.hashfile'),
                 '-hdr', path.saviour('_', *subpath)])
        assert r.returncode in (0, 22)  # nothing to deduplicate


_deduplicate = deduplicate



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

import os
import tarfile

from fingertip.util import path, temp

PREBUILD_DEPS = ['tar', 'git-core', 'rpm-build']
PREINSTALL_DEPS = [
    'ansible',
    'checkpolicy',
    'duperemove',
    'git-core',
    'nmap-ncat',
    'openssh-clients',
    'policycoreutils-python-utils',
    'python3',
    'python3-CacheControl',
    'python3-GitPython',
    'python3-cloudpickle',
    'python3-colorama',
    'python3-devel',
    'python3-fasteners',
    'python3-inotify_simple',
    'python3-lockfile',
    'python3-paramiko',
    'python3-pexpect',
    'python3-pyxdg',
    'python3-rangehttpserver',
    'python3-requests',
    'python3-requests-mock',
    'python3-ruamel-yaml',
    'qemu-img',
    'qemu-system-x86',
    'rsync',
    'util-linux',
    'xfsprogs',
]


def prepare(m, preinstall=False):
    assert hasattr(m, 'fedora')
    with m:
        pkgs = PREBUILD_DEPS + PREINSTALL_DEPS if preinstall else PREBUILD_DEPS
        m = m.apply('ansible', 'dnf', state='present', name=pkgs,
                    install_weak_deps=False)
        m.fingertip_prepared = True
    return m


def build(m, from_=None, preinstall=False):
    tarbomb = from_ if from_ and from_.endswith('.tar') else None
    if tarbomb is None:
        if from_ is None:
            fingertip_sources = path.FINGERTIP
            assert os.path.exists(os.path.join(fingertip_sources, '.copr'))
        tarbomb = temp.disappearing_file()
        with tarfile.open(tarbomb, 'w') as tf:
            tf.add(fingertip_sources, arcname='/', filter=lambda ti:
                   ti if '/redhat/' not in ti.name else None)
    if not hasattr(m, 'fingertip_prepared'):
        m = m.apply(prepare, preinstall)
    with m:
        m.ssh.upload(tarbomb, '/tmp/fingertip.tar')
        m.expiration.depend_on_a_file(tarbomb)
        m(r'''
            set -uex
            mkdir -p /tmp/fingertip/builddir
            cd /tmp/fingertip/builddir
            tar xf /tmp/fingertip.tar
            ./.copr/build-local.sh
            dnf -y builddep --setopt=install_weak_deps=False \
                /tmp/fingertip/srpms/*.rpm
            mkdir /tmp/fingertip/rpms
            rpmbuild -rb /tmp/fingertip/srpms/*.src.rpm \
                --define "_rpmdir /tmp/fingertip/rpms"
        ''')
        m.fingertip_built = True
    return m


def main(m, from_=None, preinstall=False):
    if not hasattr(m, 'fingertip_built'):
        m = m.apply(build, from_, preinstall)
    with m:
        m.ram.min = '2G'
        m('dnf -y install /tmp/fingertip/rpms/noarch/*.rpm')
        m.fingertip_installed = True
        m('useradd user', check=False)
        m('usermod -aG fingertip user')
        m('systemctl enable --now fingertip-shared-cache')
        m('fingertip-shared-cache-use user')
    return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions for fingertip: logging.
"""

import atexit
import datetime
import logging
import os
import random
import re
import sys
import threading

from fingertip.util import path, reflink


_DISABLE_LINE_WRAP = '\x1b[?7l'
_ENABLE_LINE_WRAP = '\x1b[?7h'
_ERASE = '\x1b[K'
_REWIND = '\x1b[1000D'  # hope nobody's term is wider than 1000 cols
_RESET = '\033[0m'
DEBUG = os.getenv('FINGERTIP_DEBUG') == '1'

_STRIP = re.compile(br'\x07|'  # BEL
                    br'\x1b[c7-8]|'
                    br'\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')


def strip_control_sequences(s):
    d = _STRIP.sub(b'', s).decode()
    r = repr(d)
    return r if r'\x' in r else d


class ErasingFormatter(logging.Formatter):
    _COLORS = {logging.DEBUG: '\033[0;34m',
               logging.WARNING: '\033[1;33m',
               logging.ERROR: '\033[1;31m',
               logging.CRITICAL: '\033[0;31m\033[7m'}

    def __init__(self, erasing=False, shorten_name=False):
        # Adds its own newlines, handler's .terminator is supposed to be ''
        self.erasing, self.shorten_name = erasing, shorten_name
        orig_excepthook = sys.excepthook

        def excepthook(*a):
            sys.stderr.write('\n')
            orig_excepthook(*a)
        sys.excepthook = excepthook

    def format(self, record):
        msg = record.msg.strip()
        name = record.name

        if self.shorten_name and name.startswith('fingertip.plugins.'):
            name = record.name[len('fingertip.plugins.'):]

        result = f'{name}: {msg}'

        if record.levelno in self._COLORS:
            result = self._COLORS[record.levelno] + result + _RESET

        if self.erasing:
            if record.levelno < logging.WARNING:
                pre = _ERASE + _DISABLE_LINE_WRAP
                post = _ENABLE_LINE_WRAP + _REWIND
            else:
                pre, post = _ERASE, '\n'
            return pre + result + post
        return result


class ErasingStreamHandler(logging.StreamHandler):
    def __init__(self, stream=None, erasing=True, shorten_name=True):
        super().__init__(stream)
        stream = stream or sys.stderr
        self.erasing = stream.isatty() and not DEBUG and erasing
        self.setFormatter(ErasingFormatter(erasing=self.erasing,
                                           shorten_name=shorten_name))
        if self.erasing:
            self.terminator = ''

    def stop_erasing(self):
        if self.erasing:
            sys.stderr.write(_REWIND + _ERASE)
            sys.stderr.flush()
            self.terminator = '\n'
            self.erasing = False


logger = logging.getLogger('fingertip')
critical, error, warning = logger.critical, logger.error, logger.warning
debug, info = logger.debug, logger.info
current_handler = None


def nicer():
    # global logger
    # logger = logging.getLogger('fingertip')
    global current_handler
    logger.setLevel(logging.DEBUG if DEBUG else logging.INFO)
    if current_handler:
        logger.removeHandler(current_handler)
    current_handler = ErasingStreamHandler(shorten_name=True)
    logger.addHandler(current_handler)


def plain():
    global current_handler
    logger.setLevel(logging.DEBUG if DEBUG else logging.INFO)
    if current_handler:
        if isinstance(current_handler, ErasingStreamHandler):
            current_handler.stop_erasing()
        logger.removeHandler(current_handler)
    current_handler = logging.StreamHandler()
    logger.addHandler(current_handler)


class LogPipeThread(threading.Thread):
    def __init__(self, logger, level=logging.INFO):
        threading.Thread.__init__(self, daemon=True)
        self.logger, self.level = logger, level
        self.pipe_read, self.pipe_write = os.pipe()
        self.opened_write = os.fdopen(self.pipe_write, 'wb')
        self.opened_read = os.fdopen(self.pipe_read, 'rb')
        self.opened_write.data = b''
        self.opened_write.wait = self.join
        self.start()

    def run(self):
        for line in iter(self.opened_read.readline, b''):
            self.opened_write.data += line
            if line:
                line = strip_control_sequences(line).rstrip('\r\n')
                self.logger.log(self.level, line)
        self.opened_read.close()


class LogPseudoFile():
    def __init__(self, logger, level=logging.INFO):
        self.logger, self.level = logger, level
        self._buffer = ''

    def write(self, d):
        self._buffer += d
        lines = self._buffer.split('\n')
        for line in lines[:-1]:
            if line:
                line = strip_control_sequences(line.encode()).rstrip('\r\n')
                self.logger.log(self.level, line)
        self._buffer = lines[-1]

    def flush(self):
        pass


class Sublogger:
    def __init__(self, name, to_file=None):
        # If I don't do this, it assumes that subloggers with the same name
        # are reusable, and logs stuff *somewhere*. Ugh.
        self.sub = logger.getChild(name + '.' + str(random.random()))
        self.sub.name = self.name = name
        self.path = to_file
        self.used = False

    def hint(self):
        if not os.path.exists(self.path):
            self.warning(f'{self.path} missing!')

        fname = f'{datetime.datetime.utcnow().isoformat()}.txt'
        t = path.logs(fname, makedirs=True)
        reflink.auto(self.path, t)
        home = os.path.expanduser('~')
        t = t if not t.startswith(home) else t.replace(home, '~')
        m = (f'For an intermediate log, check {t} or set FINGERTIP_DEBUG=1.'
             if not DEBUG else f'Logfile: {t}')
        sys.stderr.write(m + '\n')

    def initialize(self):
        if not self.used:
            self.used = True

            if self.path:
                self.sub.addHandler(logging.FileHandler(self.path))

                debug(f'logging to {self.path}, enabling hint')
                atexit.register(self.hint)

    def finalize(self):
        if self.used and self.path:
            debug(f'no longer logging to {self.path}, disabling hint')
            atexit.unregister(self.hint)

    def make_pipe(self, **kwargs):
        return LogPipeThread(self.sub, **kwargs).opened_write

    def pipe_powered(self, func, **what_to_supply):
        def exec_func(*a, **kwa):
            extra_args = {name: self.make_pipe(level=level)
                          for name, level in what_to_supply.items()}
            try:
                return func(*a, **kwa, **extra_args)
            finally:
                for pipe in extra_args.values():
                    pipe.close()
        return exec_func

    def make_pseudofile(self, **kwargs):
        return LogPseudoFile(self.sub, **kwargs)

    def pseudofile_powered(self, func, **what_to_supply):
        def exec_func(*a, **kwa):
            extra_args = {name: self.make_pseudofile(level=level)
                          for name, level in what_to_supply.items()}
            return func(*a, **kwa, **extra_args)
        return exec_func

    plain, nicer = staticmethod(plain), staticmethod(nicer)

    def critical(self, *args, **kwargs):
        self.initialize()
        return self.sub.critical(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.initialize()
        return self.sub.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.initialize()
        return self.sub.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.initialize()
        return self.sub.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.initialize()
        return self.sub.debug(*args, **kwargs)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


def setup(m):
    with m:
        # yes, 5 sec is a lot, but ballooning down VM RAM might be slow
        m.console.sendline("echo '>'pre-sleep; sleep 5; echo '>'post-sleep")
        m.console.expect_exact('>pre-sleep')
        return m


@fingertip.transient
def main(m):
    with m.apply(setup) as m:
        m.console.expect_exact('>post-sleep')
        m.console.expect_exact(m.prompt)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions for fingertip: finding a (hopefully free) TCP port.
"""

import socket


def find(address='127.0.0.1'):
    # race-condition-prone, but still better than just random
    # no, I don't want to pull in port-for or something
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind((address, 0))
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    _, port = s.getsockname()
    s.close()
    return port



import sys

import fingertip


@fingertip.transient
def main(m, unseal=True):
    m = m.apply('unseal') if unseal else m

    assert hasattr(m, '_backend_mode')
    if hasattr(m, 'qemu'):
        m.log.warning('^A x (could be ^A^A x in screen) or power off to exit. '
                      'Fake prompt:')
        assert not m._up_counter

        m.log.plain()

        sys.stderr.flush()
        sys.stderr.write(m.prompt)
        sys.stderr.flush()
    else:
        m.log.plain()

    m._backend_mode = 'direct'
    with m:
        sys.exit(0)  # will happen only the underlying process exits



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip.exec


def main(m, *args, unseal=True, transient=False,
         shell=True, check=True):
    m = m.apply('unseal') if unseal else m
    with m:
        r = m.exec(*(args if not shell else ('sh', '-c', ' '.join(args))))
    if not r and check:
        raise fingertip.exec.CommandExecutionError(r)
    if not transient:
        return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

from fingertip.util import log


class TimeDesync:
    """
    An abstraction to track the rough scale of time desync of the machines and:

    * invoke time synchronization (``.time_desync.fix``)
    * invoke time synchronization only if the scale of the current desync
      is too high (``.time_desync.fix_if_needed``)
    * request automatic time synchronization
      for the rest of the pipeline (``.time_desync.tighten``)

    ``NONE`` scale denotes as little-as-possible desynchronization;
    ``SMALL`` is for desyncs less then execution duration,
    caused by pauses or rewinds; and
    ``LARGE`` is for desyncs less than expiration time,
    caused by booting an old persisted snapshot.
    """
    NONE = 0
    SMALL = 1
    LARGE = 2

    @classmethod
    def _parse_scale(cls, scale):
        if isinstance(scale, str):
            return getattr(cls, scale.upper())
        return int(scale)

    def __init__(self, m):
        self._m = m
        self._current_scale = TimeDesync.NONE
        self._allowed_scale = TimeDesync.LARGE

    def report(self, scale):
        """
        Make note of a time desync of a specified scale
        """
        scale = self._parse_scale(scale)
        if scale > self._current_scale:
            log.debug(f'time desync scale increased to {scale}')
            self._current_scale = scale

    def tighten(self, allowed_scale):
        """
        Request that this and all the further stages of the pipeline
        must maintain desynchronization at or below the specified scale
        """
        allowed_scale = self._parse_scale(allowed_scale)
        if allowed_scale < self._allowed_scale:
            log.debug(f'time desync requirement tightened to {allowed_scale}')
            self._allowed_scale = allowed_scale
        self.fix_if_needed()

    def fix_if_needed(self, at_least_if_scale=None):
        """
        Perform one-off time synchronization
        if the desync is larger than the specified scale
        or the scale previously set with ``.desync.tighten``
        """
        scale = at_least_if_scale or TimeDesync.LARGE
        if min(scale, self._allowed_scale) < self._current_scale:
            log.debug('timesync '
                      f'(requested: {scale},'
                      f' allowed: {self._allowed_scale},'
                      f' current: {self._current_scale})')
            self.fix()

    def fix(self, force=False):
        """
        Perform one-off time synchronization
        :param bool force: Synchronize even if there's no detected desync
        """
        if self._current_scale != TimeDesync.NONE or force:
            assert self._m.hooks.timesync()
            self._current_scale = TimeDesync.NONE



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import hashlib
import http
import http.server
import os
import shutil
import socketserver
import stat
import threading
import time
import urllib3

import requests
import requests_mock
import cachecontrol
import cachecontrol.caches
import RangeHTTPServer

from fingertip.util import log, path, reflink


BIG = 2**30  # too big for caching
STRIP_HEADERS = ('TE', 'Transfer-Encoding', 'Keep-Alive', 'Trailer', 'Upgrade',
                 'Connection', 'Host', 'Accept')
SAVIOUR_DEFAULTS = 'local,cached+direct'
RETRIES_MAX = 7
COOLDOWN = 20


def is_cache_group_writeable():
    if os.path.exists(path.CACHE):
        mode = stat.S_IMODE(os.stat(path.CACHE).st_mode)
        return bool(mode & 0o020)


def saviour_sources():
    return [(t[len('cached+'):] if t.startswith('cached+') else t,
             t.startswith('cached+'))
            for t in
            os.getenv('FINGERTIP_SAVIOUR', SAVIOUR_DEFAULTS).split(',')]


def is_fetcheable(source, url, timeout=2):
    if source == 'local':
        return os.path.exists(path.saviour(url))
    elif source != 'direct':
        url = source + '/' + url
        url = 'http://' + url if '://' not in source else url
    try:
        r = requests.head(url, allow_redirects=False, timeout=timeout)
        return r.status_code < 400
    except (requests.exceptions.BaseHTTPError, urllib3.exceptions.HTTPError,
            requests.exceptions.Timeout, OSError) as ex:
        log.warning(f'{ex}')
        return False
    return False


class ThreadingHTTPServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    daemon_threads = True


class HTTPCache:
    def __init__(self, host='127.0.0.1', port=0):
        self.host = host
        self._mocks = {}
        self._local_files_to_serve = {}
        http_cache = self

        class Handler(RangeHTTPServer.RangeRequestHandler):
            protocol_version = 'HTTP/1.1'

            def __init__(self, *args, directory=None, **kwargs):
                super().__init__(*args, directory=path.SAVIOUR, **kwargs)

            def _status_and_headers(self, status_code, headers):
                self.send_response(status_code)
                for k, v in headers.items():
                    self.send_header(k, v)
                self.end_headers()

            def _serve(self, uri, headers, meth='GET'):
                uri = uri.lstrip('/')
                if uri in http_cache._mocks:
                    return self._serve_http(uri, headers, meth, cache=False)
                sources = saviour_sources()
                for i, (source, cache) in enumerate(sources):
                    if is_fetcheable(source, uri) or i == len(sources) - 1:
                        if source == 'local':
                            if meth == 'GET':
                                return super().do_GET()
                            elif meth == 'HEAD':
                                return super().do_HEAD()
                        elif source == 'direct':
                            su = uri
                        else:
                            su = source + '/' + uri
                            su = 'http://' + su if '://' not in source else su
                        return self._serve_http(su, headers, meth, cache=cache)

            def _serve_http(self, uri, headers, meth='GET', cache=True,
                            retries=RETRIES_MAX):
                sess = http_cache._get_requests_session(direct=not cache)
                sess_dir = http_cache._get_requests_session(direct=True)
                basename = os.path.basename(uri)

                headers = {k: v for k, v in headers.items() if
                           not (k in STRIP_HEADERS or k.startswith('Proxy-'))}
                headers['Accept-Encoding'] = 'identity'
                log.debug(f'{meth} {basename} ({uri})')
                for k, v in headers.items():
                    log.debug(f'{k}: {v}')

                error = None
                try:
                    if meth == 'GET':
                        # direct streaming or trickery might be required...
                        preview = sess.head(uri, headers=headers,
                                            allow_redirects=False)
                        if (300 <= preview.status_code < 400 and
                                'Location' in preview.headers):
                            nu = preview.headers['Location']
                            if nu.startswith('https://'):
                                # no point in serving that, we have to pretend
                                # that never happened
                                log.info(f'suppressing HTTPS redirect {nu}')
                                return self._serve_http(nu, headers, meth=meth,
                                                        cache=cache,
                                                        retries=retries)
                        direct = []
                        if not cache:
                            direct.append('caching disabled for this source')
                        if int(preview.headers.get('Content-Length', 0)) > BIG:
                            direct.append(f'file bigger than {BIG}')
                        if 'Range' in headers:
                            # There seems to be a bug in CacheControl
                            # that serves contents in full if a range request
                            # hits a non-ranged cached entry.
                            direct.append('ranged request, playing safe')
                        if direct:
                            # Don't cache, don't reencode, stream it as is
                            log.info(f'streaming {basename} directly '
                                     f'from {uri} ({", ".join(direct)})')
                            r = sess_dir.get(uri, headers=headers, stream=True)
                            self._status_and_headers(r.status_code, r.headers)
                            shutil.copyfileobj(r.raw, self.wfile)
                            return

                    # fetch with caching
                    m_func = getattr(sess, meth.lower())
                    r = m_func(uri if '://' in uri else 'http://self' + uri,
                               headers=headers, allow_redirects=False)
                    data = r.content
                    length = int(r.headers.get('Content-Length', 0))
                    if len(data) != length:
                        data = hack_around_unpacking(uri, headers, data)
                    assert len(data) == length
                except BrokenPipeError:
                    error = f'Upwards broken pipe for {meth} {uri}'
                except ConnectionResetError:
                    error = f'Upwards connection reset for {meth} {uri}'
                except requests.exceptions.ConnectionError:
                    error = f'Upwards connection error for {meth} {uri}'
                if error:
                    # delay a re-request
                    if retries:
                        log.warning(f'{error} (will retry x{retries})')
                        t = (RETRIES_MAX - retries) / RETRIES_MAX * COOLDOWN
                        time.sleep(t)
                        return self._serve_http(uri, headers, meth=meth,
                                                cache=cache,
                                                retries=retries-1)
                    else:
                        log.error(f'{error} (out of retries)')
                        self.send_error(http.HTTPStatus.SERVICE_UNAVAILABLE)
                        return
                log.debug(f'{meth} {basename} fetched {length} ({uri})')
                try:
                    self._status_and_headers(r.status_code, r.headers)
                    if meth == 'GET':
                        self.wfile.write(data)
                except BrokenPipeError:
                    log.warning(f'Downwards broken pipe for {meth} {uri}')
                except ConnectionResetError:
                    log.warning(f'Downwards connection reset for {meth} {uri}')
                except requests.exceptions.ConnectionError:
                    log.warning(f'Downwards connection error for {meth} {uri}')
                log.info(f'{meth} {basename} served {length} ({uri})')

            def do_HEAD(self):
                if self.path in http_cache._local_files_to_serve:
                    return super().do_HEAD()  # act as a HTTP server
                self._serve(uri=self.path, headers=self.headers, meth='HEAD')

            def do_GET(self):
                if self.path in http_cache._local_files_to_serve:
                    return super().do_GET()  # act as a HTTP server
                self._serve(uri=self.path, headers=self.headers, meth='GET')

            def log_message(self, format, *args):  # supress existing logging
                return

            def translate_path(self, http_path):  # directly serve local files
                if http_path in http_cache._local_files_to_serve:
                    local_path = http_cache._local_files_to_serve[http_path]
                else:
                    local_path = super().translate_path(http_path)
                log.info(f'serving {os.path.basename(http_path)} '
                         f'directly from {local_path}')
                return local_path

        httpd = ThreadingHTTPServer((host, port), Handler)
        _, self.port = httpd.socket.getsockname()
        threading.Thread(target=httpd.serve_forever, daemon=True).start()

    def _get_requests_session(self, direct=False):
        if not direct:
            kwargs = ({'filemode': 0o0660, 'dirmode': 0o0770}
                      if is_cache_group_writeable() else {})
            cache = cachecontrol.caches.FileCache(path.downloads('cache'),
                                                  **kwargs)
            sess = cachecontrol.CacheControl(requests.Session(), cache=cache)
        else:
            sess = requests.Session()
        for uri, kwargs in self._mocks.items():
            adapter = requests_mock.Adapter()
            adapter.register_uri('HEAD', uri, **kwargs)
            adapter.register_uri('GET', uri, **kwargs)
            sess.mount(uri, adapter)
        return sess

    def is_fetcheable(self, url):
        return any((is_fetcheable(src, url) for src in saviour_sources()))

    def fetch(self, url, out_path):
        sources = saviour_sources()
        for i, (source, cache) in enumerate(sources):
            if is_fetcheable(source, url) or i == len(sources) - 1:
                if source == 'local':
                    reflink.auto(path.saviour(url), out_path)
                    return
                sess = self._get_requests_session(direct=not cache)
                if source == 'direct':
                    surl = url
                else:
                    surl = source + '/' + url
                    surl = 'http://' + surl if '://' not in source else surl
                log.debug(f'fetching{"/caching" if cache else ""} '
                          f'{os.path.basename(url)} from {surl}')
                r = sess.get(surl)  # not raw because that punctures cache
                with open(out_path, 'wb') as f:
                    f.write(r.content)
                return

    def mock(self, uri, text):
        """
        Mock a text file at some location.
        Examples:
          * mock('http://self/test', text='TEST')
            and access through proxy as `http://self/test`
        """
        content_length = {'Content-Length': str(len(text.encode()))}
        self._mocks[uri] = {'text': text, 'headers': content_length}

    def mock_custom(self, uri, **kwargs):
        """
        Mock a custom HTTP response.
        See ``help(requests_mock.Adapter.register_uri)`` for parameters.
        """
        self._mocks[uri] = kwargs

    def serve_local_file(self, http_path, local_path):
        """
        Serve this local file as an HTTP server.
        Examples:
          * serve('test', '/some/file')
            and access directly as `http://<host>:<port>/test`
          * serve('http://test', '/some/file')
            and access through proxy as `http://test`
        """
        self._local_files_to_serve[http_path] = local_path


# Hack around requests uncompressing Content-Encoding: gzip


def hack_around_unpacking(uri, headers, wrong_content):
    log.warning(f're-fetching correct content for {uri}')
    r = requests.get(uri, headers=headers, stream=True, allow_redirects=False)
    h = hashlib.sha256(wrong_content).hexdigest()
    cachefile = path.downloads('fixups', h, makedirs=True)
    if not os.path.exists(cachefile):
        with path.wip(cachefile) as wip:
            with open(wip, 'wb') as f:
                shutil.copyfileobj(r.raw, f)
    with open(cachefile, 'rb') as f:
        return f.read()



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


PROMPTS = [f'{i}> ' for i in range(20)]


def apply_prompt(m, p):
    with m:
        m.console.sendline('')
        m.console.expect_exact(m.prompt)
        m.console.sendline(f'export PS1="{p}"')
        m.console.expect_exact(p)
        m.prompt = p
        return m


@fingertip.transient
def main(m, greeting='Hello!'):
    assert hasattr(m, 'console')
    for p in PROMPTS:
        m = m.apply(apply_prompt, p)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import textwrap
import selectors


class ExecResult:
    def __init__(self, retcode, out=None, err=None):
        self.retcode = retcode
        self.out, self.err = out, err

    def __bool__(self):
        return not bool(self.retcode)

    def __iter__(self):
        yield from (self.retcode, self.out, self.err)


class CommandExecutionError(RuntimeError):
    def __init__(self, exec_result):
        super().__init__(f'Command returned {exec_result.retcode}')


def nice_exec(m, *args,
              shell=True, dedent=True, check=True, decode=True):
    m.log.info(f'{args}')
    if shell and dedent:
        args = (textwrap.dedent(args[0]),)
    if shell:
        assert len(args) == 1

    exec_result = m.exec(*args, shell=shell)

    if decode:
        exec_result.out = exec_result.out.decode()
        exec_result.err = exec_result.err.decode()

    if check and not exec_result:
        m.log.error('stdout')
        m.log.error(str(exec_result.out))
        m.log.critical('stderr')
        m.log.critical(str(exec_result.err))
        raise CommandExecutionError(exec_result)

    return exec_result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fingertip',
 'fingertip.plugins',
 'fingertip.plugins.backend',
 'fingertip.plugins.os',
 'fingertip.plugins.script',
 'fingertip.plugins.self_test',
 'fingertip.plugins.software',
 'fingertip.util']

package_data = \
{'': ['*']}

install_requires = \
['CacheControl>=0.12.6,<0.13.0',
 'GitPython>=3.1.0,<4.0.0',
 'cloudpickle>=1.3.0,<2.0.0',
 'colorama>=0.4.3,<0.5.0',
 'fasteners>=0.15,<0.16',
 'inotify_simple>=1.3,<2.0',
 'lockfile>=0.12.2,<0.13.0',
 'paramiko>=2.7.1,<3.0.0',
 'pexpect>=4.8.0,<5.0.0',
 'pyxdg>=0.26,<0.27',
 'rangehttpserver>=1.2.0,<2.0.0',
 'requests-mock>=1.7.0,<2.0.0',
 'requests>=2.23.0,<3.0.0']

entry_points = \
{'console_scripts': ['fingertip = fingertip.main:main']}

setup_kwargs = {
    'name': 'fingertip',
    'version': '0.3.3',
    'description': 'Control VMs, containers and other machines with Python, leverage live snapshots',
    'long_description': '`fingertip`\n-----------\n\nThis program/library aims to be a way to:\n\n* fire up VMs and containers in mere seconds using live VM snapshots\n* uniformly control machines from Python by writing small and neat functions\n  transforming them\n* compose and transparently cache the results of these functions\n* build cool apps that exploit live VM snapshots and checkpoints\n* control other types of machines that are not local VMs\n\nAll while striving to be intentionally underengineered\nand imposing as few limits as possible.\nIf you look at it and think that it does nothing in the laziest way possible,\nthat\'s it.\n\nIt\'s currently in alpha stage.\n\n# Teaser\n\nSome examples of executing it from your shell:\n\n``` bash\n$ fingertip os.fedora + ssh  # install Fedora and SSH into it\n$ fingertip os.alpine + console  # install Alpine, access serial console\n$ fingertip os.alpine + ansible package --state=present --name=patch + ssh\n$ fingertip backend.podman-criu + os.alpine + console  # containers!\n$ fingertip os.fedora + script.debug myscript.sh  # checkpoint-powered debugger\n```\n\nAn example of Python usage and writing your own steps:\n\n``` python\nimport fingertip\n\ndef main(m=None, alias=\'itself\'):\n    m = m or fingertip.build(\'os.fedora\')\n    m = m.apply(\'ansible\', \'lineinfile\', path=\'/etc/hosts\',\n                line=f\'127.0.0.1 {alias}\')\n    with m:\n        assert \'1 received\' in m(f\'ping -c1 {alias}\').out\n    return m\n```\n\nPut in `fingertip/plugins/demo.py`,\nthis can be now be used in pipelines:\n```\n$ fingertip demo\n$ fingertip os.fedora + demo me\n$ fingertip os.alpine + demo --alias=myself + ssh\n```\n\n## Installation\n\nRefer to [INSTALL.md](INSTALL.md).\n\n\n### Shell usage\n\nIf you have installed fingertip, invoke it as `fingertip`.\n\nIf you\'re running from a checkout, use `python3 <path to checkout>` instead\nor make an alias.\n\nIf you\'re using a containerized version, invoke `fingertip-containerized`\n(and hope for the best).\n\nSo,\n\n``` bash\n$ fingertip os.fedora + ssh\n```\n\nYou should observe Fedora installation starting up,\nthen shutting down, booting up again and, finally,\ngiving you interactive control over the machine over SSH.\n\nInvoke the same command again, and it should do nearly nothing, as\nthe downloads and the installation are already cached\nin `~/.cache/fingertip`.\nEnjoy fresh clean VMs brought up in mere seconds.\nFeel like they\'re already at your fingertips.\nControl them from console or from Python.\n\n\n## Python usage\n\nLet\'s see how manipulating machines can look like\n(`fingertip/plugins/self_test/greeting.py`):\n\n``` python\ndef make_greeting(m, greeting=\'Hello!\'):                      # take a machine\n    with m:                                                   # start if needed\n        m.console.sendline(f"echo \'{greeting}\' > .greeting")  # type a command\n        m.console.expect_exact(m.prompt)                      # wait for prompt\n    return m                                                  # cache result\n\n\n@fingertip.transient                                          # don\'t lock/save\ndef main(m, greeting=\'Hello!\'):                               # take a machine\n    m = m.apply(make_greeting, greeting=greeting)             # use cached step\n    with m:                                                   # start if needed\n        assert m(\'cat .greeting\').out.strip() == greeting     # execute command\n                                                              # do not save\n```\n\n\nPlugins are regular Python functions, nothing fancy.\nYou can just pass them `fingertip.build(\'fedora\')` and that\'ll work.\nEven this `@fingertip.transient` thing\nis just an optimization hint to `.apply()`.\n\nHere\'s what can happen inside such a function:\n\n* It accepts a machine as the first argument\n  (which may be already spun up or not, you don\'t know).\n* It inspects it and applies more functions if it wants to,\n  (extra steps applied through `.apply` are cached / reused if it\'s possible).\n* Should any custom steps or changes be applied,\n  the machine must be first spun up using a `with` block (`with m as m`).\n  All modifications to the machine must happen inside that block,\n  or risk being silently undone!\n* Return the machine if the result should be cached and used for the next steps.\n  Not returning one can and usually will undo all the changes you\'ve made.\n  If you don\'t intend to save the result, don\'t return m;\n  additionally, decorate the function with `@fingertip.transient`\n  so that fingertip can apply performance optimizations and avoid locking.\n  There\'s much more to it, see `docs/on_transiency.md` for details.\n\nThe first function in the chain (or the one used in `fingertip.build`)\nwill not get a machine as the first argument.\nTo write a universal function, just use:\n``` python\ndef func(m=None):\n    m = m or fingertip.build(\'fedora\')\n    ...\n```\n\n\n## Disclaimer\n\nDue to what exactly I cache and the early stage of development,\nempty your `~/.cache/fingertip/machines` often, at least after each update.\n\n``` bash\n$ fingertip cleanup machines all\n```\n\nSome days the whole `~/.cache/fingertip` has to go.\n\n``` bash\n$ fingertip cleanup everything\n```\n',
    'author': 'Alexander Sosedkin',
    'author_email': 'asosedki@redhat.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/t184256/fingertip',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


def check_processes(m):
    if hasattr(m, 'console'):
        with m:
            # There should not be any hanging process after installation
            m.console.sendline("ps ax | grep '[s]shd.*notty' > .sshd1")
            m.console.expect_exact(m.prompt)

            if hasattr(m, 'ssh'):
                # debug output of all sshd processes
                r = m("ps ax | grep -A1 '[s]shd' && echo $$")
                assert r.retcode == 0
                m.ssh.invalidate()

            # There should not be any hanging process even after the above command
            m.console.sendline("ps ax | grep '[s]shd.*notty' > .sshd2")
            m.console.expect_exact(m.prompt)
    else:
        raise NotImplementedError()

    return m


@fingertip.transient
def main(m):
    m = m.apply('unseal')
    with m.apply(check_processes) as m:
        m.log.info(m('cat .sshd1').out.strip())
        m.log.info(m('cat .sshd2').out.strip())
        assert m('cat .sshd1 | wc -l').out.strip() == "0"
        assert m('cat .sshd2 | wc -l').out.strip() == "0"



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import importlib

from fingertip.util import log, weak_hash


def func_and_autotag(smth, *args, **kwargs):
    return load_step(smth), autotag(smth, *args, **kwargs)


def load_step(smth):
    if isinstance(smth, str):
        if smth.startswith('.') and '=' in smth:
            # this is for changing values on objects, e.g., ... + .ram.size=2G
            chain, value = smth.split('=', 1)
            chain = chain.split('.')[1:]
            return make_assigner(chain, value)
        elif smth.startswith('.'):  #
            # this is for calling methods on objects, e.g. ... + .hooks.smth
            return make_method_caller(smth.split('.')[1:])
        try:
            # try to import abc.xyz as (import fingertip.plugins.abc.xyz).main
            module = importlib.import_module('fingertip.plugins.' + smth)
            return module.main
        except (ModuleNotFoundError, AttributeError):
            # try to import abc.xyz as (import fingertip.plugins.abc).xyz
            modname, funcname = smth.rsplit('.', 1)
            module = importlib.import_module('fingertip.plugins.' + modname)
            smth = getattr(module, funcname)
    return smth


def autotag(something, *args, **kwargs):
    log.info(f'autotag in: {something} {args} {kwargs}')
    if isinstance(something, str):
        name = something if not something.startswith('.') else '_' + something
    else:
        name = something.__module__ + '.' + something.__qualname__
        assert name.startswith('fingertip.plugins.')
        name = name[len('fingertip.plugins.'):]
        if name.endswith('.__main__'):
            name = name[:len('__main__')]
    args_str = ':'.join([f'{a}' for a in args] +
                        [f'{k}={v}' for k, v in sorted(kwargs.items())])
    if args_str and (' ' in args_str or '/' in args_str or len(args_str) > 20):
        args_str = '::' + weak_hash.of_string(args_str)
    tag = f'{name}:{args_str}' if args_str else name
    return tag


def make_method_caller(name_chain):
    def method_caller(m, *args, **kwargs):
        with m:
            x = m
            for name in name_chain:
                x = getattr(x, name)
            x(*args, **kwargs)
        return m
    return method_caller


def make_assigner(name_chain, value):
    def assigner(m):
        with m:
            x = m
            for name in name_chain[:-1]:
                x = getattr(x, name)
            setattr(x, name_chain[-1], value)
        return m
    return assigner
